/*********************************************************************
 * Desktop Station air Sketch for DSair Hardware Platform
 *
 * Copyright (C) 2018 DesktopStation Co.,Ltd. / Yaasan
 * https://desktopstation.net/
 */

//#define DEBUG

// include the library code:
#include <avr/wdt.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_MM2.h"
#include "DSCoreM.h"
#include "MSensor.h"
#include "DSLCDManager.h"
#include "DSSequence.h"
#include "Functions.h"
#include "OzOLED.h"

#define FIRMWARE_VER   '5' /* 0-9, a-z, A-Z*/


#define LEDSTATE_STOP 0
#define LEDSTATE_RUN 1
#define LEDSTATE_ERR 2
#define LEDSTATE_NOSD 3
#define LEDSTATE_ANALOG 4

#define PIN_RUNLED	8	//PB0


#define THRESHOLD_EDC 276  // 1024d * (1.2 / 8) * 9V / 5V
#define THRESHOLD_OV  553  // 1024d * (1.2 / 8) * 20V / 5V


typedef struct {
	byte mOV:1;
	byte mOC:1;
	byte mLV:1;
	byte mUnk1:1;
	byte mUnk2:1;
	byte reserved:3;
} DATA_STATUS;




/* Decralation classes */
DSCoreLib DSCore;
DSSequence Sequence;
MSensor Sensor(SENSOR_CURRENT, SENSOR_VOLTAGE);
DSLCDManager LCDManager;

/* LCD functions */
OzOLED LCD;


/* LED functions */
byte gLED_State = LEDSTATE_STOP;
byte gLED_Toggle = 0;
byte gLED_Counter = 0;

/* Sensor */
byte gErrorFlag = 0;
short gOffset_Current = 0;
word gEdc_data = 0;
short gCur_data = 0;
uint8_t gSwSensor = 0;


/* Error Threshold */
word gThreshold_OV = 0;
word gThreshold_LV = 0;
byte gLastErrorNo = 0;
unsigned short gOCLevel = 80; //7.5Amax (**DSmainR5では154d**)

/* Parser */
word gLocAddresses[4] = {0,0,0,0};
byte numOfLocAddr = 0;
boolean result;

DATA_STATUS gStatus;


//Task Schedule
unsigned long gPreviousL1 = 0; //   50ms interval
unsigned long gPreviousL2 = 0; //  200ms interval
unsigned long gPreviousL3 = 0; //  500ms interval


void SetAnalogSpeed(word inSpeed, byte inDir);
void outputDammyPulse();


/***********************************************************************
* 
* Boot setup
* 
* 
************************************************************************/

void setup()
{
	
	/* OLED init */
	LCD.OutputDammy = outputDammyPulse;
	LCD.init();
	LCD.clearDisplay();           //clear the screen and set start position to top left corner
	LCD.setNormalDisplay();       //Set display to Normal mode
	LCD.setPageMode();            //Set addressing mode to Page Mode
	LCD.printString("DS", 0, 3);
	LCD.printString("POWER HACX", 0, 2);
	LCD.printString("R4 FW1", 0, 5);
	
	delay(500);
	
	/* Debug message for PC */
	Serial.begin(115200);
	while (!Serial);
	
	//Serial.println(F("DS HACX R4 FW1"));
	
	//Pin mode
	pinMode(PIN_PWMA, OUTPUT);
	pinMode(PIN_PWMB, OUTPUT);
	
	pinMode(PIN_RUNLED, OUTPUT);
	
	//Register Event functions
	Sequence.ChangedTargetLocAddr = changedTargetLocAddr;
	Sequence.ChangedTargetSpeed = changedTargetSpeed;
	Sequence.ChangedTargetFunc = changedTargetFunc;
	Sequence.ChangedTargetDir = changedTargetDir;
	Sequence.ChangedTurnout = changedTurnout;
	Sequence.ChangedPowerStatus = changedPowerStatus;
	Sequence.DebugMessage = debugMessage;
	Sequence.ChangedMain = changedMain;
	Sequence.ChangedMenu = changedMenu;
	Sequence.ChangedAnalog = changedAnalog;
	Sequence.ChangedProtocol = changedProtocol;
	Sequence.DisplayClear = clearDisplay;
	Sequence.ReadCVFunc = readCVFunc;
	Sequence.WriteCVFunc = writeCVFunc;
	
	
	LCDManager.PrintString = printString;
	LCDManager.PrintStringNOT = printStringNOT;
	LCDManager.PrintStringBig = printStringBig;
	LCDManager.ClearDisplay = clearDisplay;
	LCDManager.DrawIcon = drawIcon;
	LCDManager.DrawSpeedBar = drawSpeedBar;
	
	
	
	/* DSCore初期化 */
	DSCore.Init();
	
	/* 線路に電源遮断 */
	DS_Power(0);
	
	DSCore.gCutOut = 0;//RailCom Default Off
	
	//Dammy Registration(Addr.3)
	DSCore.SetLocoSpeedEx(0xC000 + 3, 0, 2);
	
	
	/* LED turn on */
	changeLED();
	
	//Serial.println(F("100 Ready"));
	
	//Reset task
	gPreviousL1 = millis();
	gPreviousL2 = millis();
	gPreviousL3 = millis();
	
	//Timer1 490Hz -> 31.4kHz
	TCCR1B &= B11111000;
	TCCR1B |= B00000001;	
	
	//Clear Display
	LCD.lcd_clear();
	LCDManager.RegisterLocDirection(Sequence.CurrentProtocol + Sequence.NextLocoAddress, Sequence.NextLocoDirection);
	LCDManager.RegisterMain();
	LCDManager.Interval();
	
	
	/* Dammy sleep */
	Sensor.CalcCurrentOffset();
	
	sei();    //ENABLE INTERRUPTION
	

	
}





/***********************************************************************
* 
* Timer Interval
* 
************************************************************************/

void loop()
{
	
	//パルスジェネレータのスキャン
	DSCore.Scan();
	
	if( (millis() - gPreviousL1) >= 100)
	{
		//Reset task
		gPreviousL1 = millis();
		
		//Sequence
		Sequence.Interval_Fast();
		Sequence.Interval_Slow();
		
	}
	
	if( (millis() - gPreviousL2) >= 200)
	{
		//Reset task
		gPreviousL2 = millis();
		
		//LED Control
		taskLED();
		
	}
	
	if( (millis() - gPreviousL3) >= 500)
	{
		//Reset task
		gPreviousL3 = millis();
		
		//Monitor Current
		Sensor.MonitorCurrent();
	
		/* Sensor updates */
		unsigned char aErrorFlag = Sensor.Update(gOCLevel);
		
		/*
		if( aErrorFlag > 0)
		{
			
			//電源供給中の時
			if( Sequence.NextPowerStatus > 0)
			{
				Sequence.NextPowerStatus = 0;
				DSCore.SetPower(0);
				
				LCDManager.SetSensorsErr(Sensor.GetCurrent(), Sensor.GetVoltage());
				
				LCDManager.RegisterPower(0);
				
				Sequence.NextMode(SQCMD_DEPTH_INFO);
				LCDManager.RegisterInfo();
				
			}
			
		}
		else
		{
			
		}
		*/
		
		printSensors();
		
		/* 表示 */
		LCDManager.Interval();	
	
	
	}
	

}

void printSensors()
{
	unsigned char aFlagDisplay = gErrorFlag;
	
	unsigned short aCurrent = Sensor.GetCurrent();
	unsigned char aVoltage = Sensor.GetVoltage();
	
	unsigned short aCurrent_H = aCurrent / 100;
	unsigned short aCurrent_L = aCurrent - (aCurrent_H * 100);
	unsigned char aVoltage_H = aVoltage / 10;
	unsigned char aVoltage_L = aVoltage - (aVoltage_H * 10);
	
	//情報表示の時は更新する
	LCDManager.SetSensors(aCurrent_H, aCurrent_L, aVoltage_H, aVoltage_L, aFlagDisplay);
	
	
}


void PowerOffByErr(uint8_t inErrNo)
{
	ReplyPowerPacket(0);
	DSCore.SetPower(0);
	
	/* 状態変更 */
	gLED_State = LEDSTATE_ERR;
	gLED_Counter =0;
	gLastErrorNo = inErrNo;
	
}

void ReplyPowerPacket(byte inPower)
{
	#ifdef DEBUG
	/* Power */
	Serial.print("@PWR,");
	Serial.print(inPower);
	Serial.println(",");
	#endif
}


void changeLED(void)
{
	
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		digitalWrite(PIN_RUNLED, 1);
		break;
		
	case LEDSTATE_RUN:
		digitalWrite(PIN_RUNLED, 0);
		break;
	
	case LEDSTATE_ANALOG:
		break;
		
	case LEDSTATE_ERR:
		break;
		
	case LEDSTATE_NOSD:
		break;
	}
}

void taskLED(void)
{
	
	/* カウンタを回す(200msで1カウント) */
	gLED_Counter++;
	gLED_Counter = gLED_Counter & 0b1111;
	
	//LED操作
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		
		if( gLED_Counter <= 3)
		{
			digitalWrite(PIN_RUNLED, 0);
		}
		else
		{
			digitalWrite(PIN_RUNLED, 1);
		}
		break;
		
	case LEDSTATE_RUN:
		break;
		
	case LEDSTATE_ERR:
		
		if( gLED_Toggle == 0)
		{
			digitalWrite(PIN_RUNLED, 0);
		}
		else
		{
			digitalWrite(PIN_RUNLED, 1);
		}
		
		gLED_Toggle = (~gLED_Toggle) & 1;
		break;
	
	case LEDSTATE_ANALOG:
		if( (gLED_Counter == 0) || (gLED_Counter == 8))
		{
			digitalWrite(PIN_RUNLED, 1);
		}
		else
		{
			digitalWrite(PIN_RUNLED, 0);
		}
		
		break;
		
	case LEDSTATE_NOSD:
		
		if( gLED_Counter <= 0)
		{
			digitalWrite(PIN_RUNLED, 1);
		}
		else if(gLED_Counter == 2)
		{
			digitalWrite(PIN_RUNLED, 1);
		}
		else
		{
			digitalWrite(PIN_RUNLED, 0);
		}
		
		break;
	}
}

void SetAnalogSpeed(word inSpeed, byte inDir)
{
	
	word aSpeed = inSpeed >> 2;
	
	if( aSpeed > 255)
	{
		aSpeed = 255;
	}
	
	if( inSpeed == 0)
	{
		FreeAnalogSpeed();
	}
	else
	{
		if( inDir == 2)
		{
			//REV
			analogWrite(PIN_PWMA, 255 - aSpeed);
			analogWrite(PIN_PWMB, 255);
		}
		else
		{
			//FWD
			analogWrite(PIN_PWMB, 255 - aSpeed);
			analogWrite(PIN_PWMA, 255);
		}
	}
}

void FreeAnalogSpeed()
{
	analogWrite(PIN_PWMB, 0);
	analogWrite(PIN_PWMA, 0);
}

void changedTargetLocAddr(const word inAddr, const byte inDir, const long inFuncBuf)
{
	LCDManager.RegisterLoc(inAddr, inDir, inFuncBuf);
	LCDManager.Interval();
}

void changedTargetSpeed(const word inAddr, const word inSpeed, const bool inUpdate)
{
	byte aProtocol = 2;
	
	if( inAddr <= 255)
	{
		aProtocol = 0;
	}
	
	if( inUpdate == true)
	{
		DSCore.SetLocoSpeedEx(inAddr, inSpeed, aProtocol);
	}
	
	LCDManager.RegisterLocSpeed(inAddr, inSpeed);
	LCDManager.Interval();

}

void changedTargetFunc(const word inAddr, const byte inFuncNo, const byte inFuncPower, const bool inUpdate, const byte inMode, const long inFuncBuf)
{
	if( inMode == 0)
	{
		//Serial.println(inFuncBuf, BIN);
		LCDManager.RegisterLocFunctionNormal(inAddr, inFuncBuf);//DEBUG
		
	}
	else
	{
		if( inUpdate == true)
		{
			DSCore.SetLocoFunction(inAddr, inFuncNo + 1, inFuncPower);
		}
		
		LCDManager.RegisterLocFunction(inAddr, inFuncNo, inFuncPower);
	}
	
	LCDManager.Interval();

}

void changedTargetDir(const word inAddr, const byte inDir, const bool inUpdate)
{
	
	if( inUpdate == true)
	{
		DSCore.SetLocoDirection(inAddr, inDir);
	}
	
	LCDManager.RegisterLocDirection(inAddr, inDir);
	LCDManager.Interval();

}

void changedTurnout(const word inAddr, const byte inDir, const bool inUpdate)
{
	if( inUpdate == true)
	{
		DSCore.SetTurnout(inAddr, inDir);
	}
	
	LCDManager.RegisterTurnout(inAddr, inDir);
	LCDManager.Interval();

}


void changedPowerStatus(const byte inPower)
{

	if( inPower == 1)
	{
		//端子を解放
		FreeAnalogSpeed();
		
		//電流オフセット計測
		Sensor.CalcCurrentOffset();
		
		/* 状態変更 */
		gLED_State = LEDSTATE_RUN;
		
		Serial.println("PWR ON");
		
	}
	else
	{
		/* 状態変更 */
		gLED_State = LEDSTATE_STOP;
		
		Serial.println("PWR OFF");
	}
	
	LCDManager.RegisterPower(inPower);
	LCDManager.Interval();
	
	/* LED状態の変更 */
	gLED_Counter = 0;
	changeLED();
	
	DSCore.SetPower(inPower);
	
}

void changedMain()
{
	LCDManager.WriteLocAddr(Sequence.CurrentProtocol + Sequence.NextLocoAddress);
	LCDManager.WriteAccAddr(Sequence.CurrentACCProtocol + Sequence.NextAccAddress);
	
	LCDManager.RegisterMain();
	LCDManager.Interval();
}


void changedMenu(byte inMenuNo)
{
	
	if( gLED_State == LEDSTATE_ANALOG)
	{
		//アナログ時は強制終了
		changedPowerStatus(0);
	}
	
	LCDManager.RegisterMenu(inMenuNo);
	LCDManager.Interval();
}

void changedAnalog()
{
	//強制DCC電源オフ
	changedPowerStatus(0);
	
	//アナログモード
	gLED_State = LEDSTATE_ANALOG;
	
	clearDisplay();
	
	//アナログ画面の表示
	LCDManager.RegisterAnalog();
	LCDManager.Interval();
}

void changedProtocol(byte inProtocol)
{
	LCDManager.RegisterProtocol(inProtocol);
}


void readCVFunc(const byte inSeq, const word inCVAddr, const byte inCVValue)
{
	byte aValue = 0;
	bool aResult;
	
	if( inSeq == 1)
	{
		LCDManager.RegisterCVRead(10, inCVAddr, 0);
		LCDManager.Interval();
		
		aValue = 0;
		aResult = DSCore.ReadConfig(inCVAddr, &aValue, 0, 0);
		
		Serial.print("CV Read: Adr.");
		Serial.print(inCVAddr);
		Serial.print(", Val.");
		Serial.print(aValue);
		Serial.println("");
		
		//Result
		if( aResult != true)
		{
			LCDManager.RegisterCVRead(11, inCVAddr, aValue);
		}
		else
		{
			LCDManager.RegisterCVRead(1, inCVAddr, aValue);
		}
		
		LCDManager.Interval();
	}
	else
	{
		LCDManager.RegisterCVRead(inSeq, inCVAddr, 0);
		LCDManager.Interval();
	}

	
}

void writeCVFunc(const byte inSeq, const word inCVAddr, const byte inCVValue)
{
	unsigned long aBaseTime = 0;
	word aCurOffset;
	bool aSuccessWrite = false;
	int aCurrent;
	
	if( inSeq == 2)
	{
		LCDManager.RegisterCVWrite(20, inCVAddr, inCVValue);
		LCDManager.Interval();
		
		
		DSCore.WriteConfig_Dir(inCVAddr, inCVValue);
		
		//電流オフセットを取得
		aCurOffset = Sensor.GetCurrentOffset();
		
		//5secのタイムアウト or 電流 60mA待ち
		aBaseTime = millis();
		
		while((millis() - aBaseTime) < 5000)
		{
			aCurrent = (int)analogRead(SENSOR_CURRENT) - (int)aCurOffset;
			
			//SerialDS.printDEC(aCurrent);
			//SerialDS.print(",");
			
			if( aCurrent > 10)
			{
				//成功
				aSuccessWrite = true;
				
				break;
			}
			
			
		}
		//delay(2000);
		
		
		LCDManager.RegisterCVWrite(inSeq, aSuccessWrite == true ? inCVAddr : 0xFFFF, inCVValue);
		LCDManager.Interval();
		
	}
	else
	{
		LCDManager.RegisterCVWrite(inSeq, inCVAddr, inCVValue);
		LCDManager.Interval();
	}
	
}

void debugMessage(const byte inPower)
{
	Serial.print("DebugSeq: ");
	Serial.print(inPower);
	Serial.println("");

}

void printString(const char *s, int x, int y)
{
	LCD.printString(s, x, y);
}

void printStringNOT(const char *s, int x, int y)
{
	LCD.printStringNOT(s, x, y);
}


void printStringBig(const char *s, int x, int y)
{
	
	//LCD.setCursorXY(x, y);
	LCD.printBigNumberX2(s, x, y);
}

void drawIcon(byte x, byte y, byte num, byte inNot)
{
	LCD.drawIcon(x, y, num, inNot);
	
}

void drawSpeedBar(byte x, byte y, byte speed, byte inNot)
{
	LCD.drawSpeedBar(x, y, speed, false);
}

void clearDisplay()
{
	LCD.lcd_clear();           //clear the screen and set start position to top left corner

}

void outputDammyPulse()
{
	DSCore.TogglePulse();
}

